"""
Computing the feature vectors is the most computationally expensive part of creating these
timelapses. This module is responsible for offloading that computation to the GPU to improve
throughput.
"""
